// Copyright © 2016 Laurence Gonsalves
//
// This file is part of kotlin-argparser, a library which can be found at
// http://github.com/xenomachina/kotlin-argparser
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published by the
// Free Software Foundation; either version 2.1 of the License, or (at your
// option) any later version.
//
// This library is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
// for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this library; if not, see http://www.gnu.org/licenses/

package com.xenomachina.argparser

import com.xenomachina.common.orElse
import org.junit.Assert.assertEquals
import org.junit.Assert.assertFalse
import org.junit.Assert.assertTrue
import org.junit.Rule
import org.junit.Test
import org.junit.rules.TestName
import java.io.StringWriter

val TEST_HELP = "test help message"

class ArgParserTest {
    @JvmField @Rule
    val testName: TestName = TestName()

    inline fun <reified X : Throwable> shouldThrow(f: () -> Unit): X {
        val javaClass = X::class.java
        try {
            f()
        } catch (exception: Throwable) {
            if (javaClass.isInstance(exception)) return javaClass.cast(exception)
            throw exception
        }
        throw AssertionError("Expected ${javaClass.canonicalName} to be thrown")
    }

    fun parserOf(
            vararg args: String,
            mode: ArgParser.Mode = ArgParser.Mode.GNU,
            helpFormatter: HelpFormatter? = DefaultHelpFormatter()
    ) = ArgParser(args, mode, helpFormatter)

    @Test
    fun testArglessShortOptions() {
        class Args(parser: ArgParser) {
            val xyz by parser.option<MutableList<String>>("-x", "-y", "-z",
                    errorName = "VALUE_NAME",
                    help = TEST_HELP,
                    usageArgument = "ARG_NAME") {
                value.orElse { mutableListOf<String>() }.apply {
                    add("$optionName")
                }
            }
        }

        assertEquals(
                listOf("-x", "-y", "-z", "-z", "-y"),
                Args(parserOf("-x", "-y", "-z", "-z", "-y")).xyz)

        assertEquals(
                listOf("-x", "-y", "-z"),
                Args(parserOf("-xyz")).xyz)
    }

    @Test
    fun testShortOptionsWithArgs() {
        class Args(parser: ArgParser) {
            val a by parser.flagging("-a", help = TEST_HELP)
            val b by parser.flagging("-b", help = TEST_HELP)
            val c by parser.flagging("-c", help = TEST_HELP)
            val xyz by parser.option<MutableList<String>>("-x", "-y", "-z",
                    errorName = "VALUE_NAME",
                    usageArgument = "ARG_NAME", help = TEST_HELP) {
                value.orElse { mutableListOf<String>() }.apply {
                    add("$optionName:${next()}")
                }
            }
        }

        // Test with value as separate arg
        assertEquals(
                listOf("-x:0", "-y:1", "-z:2", "-z:3", "-y:4"),
                Args(parserOf("-x", "0", "-y", "1", "-z", "2", "-z", "3", "-y", "4")).xyz)

        // Test with value concatenated
        assertEquals(
                listOf("-x:0", "-y:1", "-z:2", "-z:3", "-y:4"),
                Args(parserOf("-x0", "-y1", "-z2", "-z3", "-y4")).xyz)

        // Test with = between option and value. Note that the "=" is treated as part of the option value for short options.
        assertEquals(
                listOf("-x:=0", "-y:=1", "-z:=2", "-z:=3", "-y:=4"),
                Args(parserOf("-x=0", "-y=1", "-z=2", "-z=3", "-y=4")).xyz)

        // Test chained options. Note that an option with arguments must be last in the chain
        val chain1 = Args(parserOf("-abxc"))
        assertTrue(chain1.a)
        assertTrue(chain1.b)
        assertFalse(chain1.c)
        assertEquals(listOf("-x:c"), chain1.xyz)

        val chain2 = Args(parserOf("-axbc"))
        assertTrue(chain2.a)
        assertFalse(chain2.b)
        assertFalse(chain2.c)
        assertEquals(listOf("-x:bc"), chain2.xyz)
    }

    @Test
    fun testMixedShortOptions() {
        class Args(parser: ArgParser) {
            val def by parser.option<MutableList<String>>("-d", "-e", "-f",
                    errorName = "VALUE_NAME",
                    usageArgument = "ARG_NAME",
                    help = TEST_HELP) {
                value.orElse { mutableListOf<String>() }.apply {
                    add("$optionName")
                }
            }
            val abc by parser.option<MutableList<String>>("-a", "-b", "-c",
                    errorName = "VALUE_NAME",
                    usageArgument = "ARG_NAME",
                    help = TEST_HELP) {
                value.orElse { mutableListOf<String>() }.apply {
                    add("$optionName")
                }
            }
        }

        Args(parserOf("-adbefccbafed")).run {
            assertEquals(
                    listOf("-d", "-e", "-f", "-f", "-e", "-d"),
                    def)
            assertEquals(
                    listOf("-a", "-b", "-c", "-c", "-b", "-a"),
                    abc)
        }
    }

    @Test
    fun testMixedShortOptionsWithArgs() {
        class Args(parser: ArgParser) {
            val def by parser.option<MutableList<String>>("-d", "-e", "-f",
                    errorName = "VALUE_NAME",
                    usageArgument = "ARG_NAME",
                    help = TEST_HELP) {
                value.orElse { mutableListOf<String>() }.apply {
                    add("$optionName")
                }
            }
            val abc by parser.option<MutableList<String>>("-a", "-b", "-c",
                    errorName = "VALUE_NAME",
                    usageArgument = "ARG_NAME",
                    help = TEST_HELP) {
                value.orElse { mutableListOf<String>() }.apply {
                    add("$optionName")
                }
            }
            val xyz by parser.option<MutableList<String>>("-x", "-y", "-z",
                    errorName = "VALUE_NAME",
                    usageArgument = "ARG_NAME",
                    help = TEST_HELP) {
                value.orElse { mutableListOf<String>() }.apply {
                    add("$optionName:${next()}")
                }
            }
        }

        Args(parserOf("-adecfy5", "-x0", "-bzxy")).run {
            assertEquals(
                    listOf("-a", "-c", "-b"),
                    abc)
            assertEquals(
                    listOf("-d", "-e", "-f"),
                    def)
            assertEquals(
                    listOf("-y:5", "-x:0", "-z:xy"),
                    xyz)
        }
    }

    @Test
    fun testArglessLongOptions() {
        class Args(parser: ArgParser) {
            val xyz by parser.option<MutableList<String>>("--xray", "--yellow", "--zebra",
                    errorName = "ARG_NAME",
                    usageArgument = null,
                    help = TEST_HELP) {
                value.orElse { mutableListOf<String>() }.apply {
                    add("$optionName")
                }
            }
        }

        assertEquals(
                listOf("--xray", "--yellow", "--zebra", "--zebra", "--yellow"),
                Args(parserOf("--xray", "--yellow", "--zebra", "--zebra", "--yellow")).xyz)

        assertEquals(
                listOf("--xray", "--yellow", "--zebra"),
                Args(parserOf("--xray", "--yellow", "--zebra")).xyz)
    }

    @Test
    fun testLongOptionsWithArgs() {
        class Args(parser: ArgParser) {
            val xyz by parser.option<MutableList<String>>("--xray", "--yellow", "--zaphod",
                    errorName = "ARG_NAME", usageArgument = "ARG_NAME",
                    help = TEST_HELP) {
                value.orElse { mutableListOf<String>() }.apply {
                    add("$optionName:${next()}")
                }
            }
        }

        // Test with value as separate arg
        assertEquals(
                listOf("--xray:0", "--yellow:1", "--zaphod:2", "--zaphod:3", "--yellow:4"),
                Args(parserOf("--xray", "0", "--yellow", "1", "--zaphod", "2", "--zaphod", "3", "--yellow", "4")).xyz)

        // Test with = between option and value
        assertEquals(
                listOf("--xray:0", "--yellow:1", "--zaphod:2", "--zaphod:3", "--yellow:4"),
                Args(parserOf("--xray=0", "--yellow=1", "--zaphod=2", "--zaphod=3", "--yellow=4")).xyz)

        shouldThrow<UnrecognizedOptionException> {
            Args(parserOf("--xray0", "--yellow1", "--zaphod2", "--zaphod3", "--yellow4")).xyz
        }.run {
            assertEquals("unrecognized option '--xray0'", message)
        }
    }

    @Test
    fun testDefault() {
        class Args(parser: ArgParser) {
            val x by parser.storing("-x",
                    help = TEST_HELP) { toInt() }.default(5)
        }

        // Test with no value
        assertEquals(
                5,
                Args(parserOf()).x)

        // Test with value
        assertEquals(
                6,
                Args(parserOf("-x6")).x)

        // Test with value as separate arg
        assertEquals(
                7,
                Args(parserOf("-x", "7")).x)

        // Test with multiple values
        assertEquals(
                8,
                Args(parserOf("-x9", "-x8")).x)
    }
    @Test
    fun testDefaultWithProvider() {
        class Args(parser: ArgParser) {
            val x by parser.storing(help = TEST_HELP) { toInt() }.default(5)
        }

        // Test with no value
        assertEquals(
                5,
                Args(parserOf()).x)

        // Test with value
        assertEquals(
                6,
                Args(parserOf("-x6")).x)

        // Test with value as separate arg
        assertEquals(
                7,
                Args(parserOf("-x", "7")).x)

        // Test with multiple values
        assertEquals(
                8,
                Args(parserOf("-x9", "-x8")).x)
    }

    @Test
    fun testFlag() {
        class Args(parser: ArgParser) {
            val x by parser.flagging("-x", "--ecks",
                    help = TEST_HELP)
            val y by parser.flagging("-y",
                    help = TEST_HELP)
            val z by parser.flagging("--zed",
                    help = TEST_HELP)
        }

        Args(parserOf("-x", "-y", "--zed", "--zed", "-y")).run {
            assertTrue(x)
            assertTrue(y)
            assertTrue(z)
        }

        Args(parserOf()).run {
            assertFalse(x)
            assertFalse(y)
            assertFalse(z)
        }

        Args(parserOf("-y", "--ecks")).run {
            assertTrue(x)
            assertTrue(y)
        }

        Args(parserOf("--zed")).run {
            assertFalse(x)
            assertFalse(y)
            assertTrue(z)
        }
    }

    @Test
    fun testArgument_noParser() {
        class Args(parser: ArgParser) {
            val x by parser.storing("--ecks", "-x",
                    help = TEST_HELP)
        }

        assertEquals("foo",
                Args(parserOf("-x", "foo")).x)

        assertEquals("baz",
                Args(parserOf("-x", "bar", "-x", "baz")).x)

        assertEquals("short",
                Args(parserOf("--ecks", "long", "-x", "short")).x)

        assertEquals("long",
                Args(parserOf("-x", "short", "--ecks", "long")).x)

        val args = Args(parserOf())
        shouldThrow<MissingValueException> {
            args.x
        }.run {
            assertEquals("missing ECKS", message)
        }
    }

    @Test
    fun testArgument_missing_long() {
        class Args(parser: ArgParser) {
            val x by parser.storing("--ecks",
                    help = TEST_HELP)
        }

        val args = Args(parserOf())
        shouldThrow<MissingValueException> {
            args.x
        }.run {
            assertEquals("missing ECKS", message)
        }
    }

    @Test
    fun testArgument_missing_short() {
        class Args(parser: ArgParser) {
            val x by parser.storing("-x",
                    help = TEST_HELP)
        }

        val args = Args(parserOf())
        shouldThrow<MissingValueException> {
            args.x
        }.run {
            assertEquals("missing X", message)
        }
    }

    @Test
    fun testArgument_withParser() {
        class Args(parser: ArgParser) {
            val x by parser.storing("-x", "--ecks",
                    help = TEST_HELP) { toInt() }
        }

        val opts1 = Args(parserOf("-x", "5"))
        assertEquals(5, opts1.x)

        val opts2 = Args(parserOf("-x", "1", "-x", "2"))
        assertEquals(2, opts2.x)

        val opts3 = Args(parserOf("--ecks", "3", "-x", "4"))
        assertEquals(4, opts3.x)

        val opts4 = Args(parserOf("-x", "5", "--ecks", "6"))
        assertEquals(6, opts4.x)

        val opts6 = Args(parserOf())
        shouldThrow<MissingValueException> {
            opts6.x
        }.run {
            assertEquals("missing ECKS", message)
        }
    }

    @Test
    fun testAccumulator_noParser() {
        class Args(parser: ArgParser) {
            val x by parser.adding("-x", "--ecks",
                    help = TEST_HELP)
        }

        assertEquals(
                listOf<String>(),
                Args(parserOf()).x)

        assertEquals(
                listOf("foo"),
                Args(parserOf("-x", "foo")).x)

        assertEquals(
                listOf("bar", "baz"),
                Args(parserOf("-x", "bar", "-x", "baz")).x)

        assertEquals(
                listOf("long", "short"),
                Args(parserOf("--ecks", "long", "-x", "short")).x)

        assertEquals(
                listOf("short", "long"),
                Args(parserOf("-x", "short", "--ecks", "long")).x)
    }

    @Test
    fun testAccumulator_withParser() {
        class Args(parser: ArgParser) {
            val x by parser.adding("-x", "--ecks",
                    help = TEST_HELP) { toInt() }
        }

        assertEquals(listOf<Int>(), Args(parserOf()).x)
        assertEquals(listOf(5), Args(parserOf("-x", "5")).x)
        assertEquals(listOf(1, 2), Args(parserOf("-x", "1", "-x", "2")).x)
        assertEquals(listOf(3, 4), Args(parserOf("--ecks", "3", "-x", "4")).x)
        assertEquals(listOf(5, 6), Args(parserOf("-x", "5", "--ecks", "6")).x)
    }

    enum class Color { RED, GREEN, BLUE }

    class ColorArgs(parser: ArgParser) {
        val color by parser.mapping(
                "--red" to Color.RED,
                "--green" to Color.GREEN,
                "--blue" to Color.BLUE,
                help = TEST_HELP)
    }

    @Test
    fun testMapping() {
        assertEquals(Color.RED, ColorArgs(parserOf("--red")).color)
        assertEquals(Color.GREEN, ColorArgs(parserOf("--green")).color)
        assertEquals(Color.BLUE, ColorArgs(parserOf("--blue")).color)

        // Last one takes precedence
        assertEquals(Color.RED, ColorArgs(parserOf("--blue", "--red")).color)
        assertEquals(Color.GREEN, ColorArgs(parserOf("--blue", "--green")).color)
        assertEquals(Color.BLUE, ColorArgs(parserOf("--red", "--blue")).color)

        val args = ColorArgs(parserOf())
        shouldThrow<MissingValueException> {
            args.color
        }.run {
            assertEquals("missing --red|--green|--blue", message)
        }
    }

    class OptionalColorArgs(parser: ArgParser) {
        val color by parser.mapping(
                "--red" to Color.RED,
                "--green" to Color.GREEN,
                "--blue" to Color.BLUE,
                help = TEST_HELP)
                .default(Color.GREEN)
    }

    @Test
    fun testMapping_withDefault() {
        assertEquals(Color.RED, OptionalColorArgs(parserOf("--red")).color)
        assertEquals(Color.GREEN, OptionalColorArgs(parserOf("--green")).color)
        assertEquals(Color.BLUE, OptionalColorArgs(parserOf("--blue")).color)
        assertEquals(Color.GREEN, OptionalColorArgs(parserOf()).color)
    }

    @Test
    fun testUnrecognizedShortOpt() {
        shouldThrow<UnrecognizedOptionException> {
            OptionalColorArgs(parserOf("-x")).color
        }.run {
            assertEquals("unrecognized option '-x'", message)
        }
    }

    @Test
    fun testUnrecognizedLongOpt() {
        shouldThrow<UnrecognizedOptionException> {
            OptionalColorArgs(parserOf("--ecks")).color
        }.run {
            assertEquals("unrecognized option '--ecks'", message)
        }
    }

    @Test
    fun testStoringNoArg() {
        class Args(parser: ArgParser) {
            val x by parser.storing("-x", "--ecks",
                    help = TEST_HELP)
        }

        // Note that name actually used for option is used in message
        shouldThrow<OptionMissingRequiredArgumentException> {
            Args(parserOf("-x")).x
        }.run {
            assertEquals("option '-x' is missing a required argument", message)
        }

        // Note that name actually used for option is used in message
        shouldThrow<OptionMissingRequiredArgumentException> {
            Args(parserOf("--ecks")).x
        }.run {
            assertEquals("option '--ecks' is missing a required argument", message)
        }
    }

    @Test
    fun testShortStoringNoArgChained() {
        class Args(parser: ArgParser) {
            val y by parser.flagging("-y",
                    help = TEST_HELP)
            val x by parser.storing("-x",
                    help = TEST_HELP)
        }

        // Note that despite chaining, hyphen appears in message
        shouldThrow<OptionMissingRequiredArgumentException> {
            Args(parserOf("-yx")).x
        }.run {
            assertEquals("option '-x' is missing a required argument", message)
        }
    }

    @Test
    fun testInitValidation() {
        class Args(parser: ArgParser) {
            val yDelegate = parser.storing("-y",
                    help = TEST_HELP) { toInt() }
            val y by yDelegate

            val xDelegate = parser.storing("-x",
                    help = TEST_HELP) { toInt() }
            val x by xDelegate

            init {
                if (y >= x)
                    throw InvalidArgumentException("${yDelegate.errorName} must be less than ${xDelegate.errorName}")

                // A better way to accomplish validation that only depends on one Delegate is to use
                // Delegate.addValidator. See testAddValidator for an example of this.
                if (x.rem(2) != 0)
                    throw InvalidArgumentException("${xDelegate.errorName} must be even, $x is odd")
            }
        }

        // This should pass validation
        val opts0 = Args(parserOf("-y1", "-x10"))
        assertEquals(1, opts0.y)
        assertEquals(10, opts0.x)

        shouldThrow<InvalidArgumentException> {
            Args(parserOf("-y20", "-x10")).x
        }.run {
            assertEquals("Y must be less than X", message)
        }

        shouldThrow<InvalidArgumentException> {
            Args(parserOf("-y10", "-x15")).x
        }.run {
            assertEquals("X must be even, 15 is odd", message)
        }

        shouldThrow<InvalidArgumentException> {
            Args(parserOf("-y10", "-x15")).x
        }.run {
            assertEquals("X must be even, 15 is odd", message)
        }
    }

    @Test
    fun testAddValidator() {
        class Args(parser: ArgParser) {
            val yDelegate = parser.storing("-y",
                    help = TEST_HELP) { toInt() }
            val y by yDelegate

            val xDelegate = parser.storing("-x",
                    help = TEST_HELP) { toInt() }
                    .addValidator {
                        if (value.rem(2) != 0)
                            throw InvalidArgumentException("$errorName must be even, $value is odd")
                    }
            val x by xDelegate

            init {
                if (y >= x)
                    throw InvalidArgumentException("${yDelegate.errorName} must be less than ${xDelegate.errorName}")
            }
        }

        // This should pass validation
        val opts0 = Args(parserOf("-y1", "-x10"))
        assertEquals(1, opts0.y)
        assertEquals(10, opts0.x)

        shouldThrow<InvalidArgumentException> {
            Args(parserOf("-y20", "-x10")).x
        }.run {
            assertEquals("Y must be less than X", message)
        }

        shouldThrow<InvalidArgumentException> {
            Args(parserOf("-y10", "-x15")).x
        }.run {
            assertEquals("X must be even, 15 is odd", message)
        }

        shouldThrow<InvalidArgumentException> {
            Args(parserOf("-y10", "-x15")).x
        }.run {
            assertEquals("X must be even, 15 is odd", message)
        }
    }

    @Test
    fun testUnconsumed() {
        class Args(parser: ArgParser) {
            val y by parser.flagging("-y", "--why",
                    help = TEST_HELP)
            val x by parser.flagging("-x", "--ecks",
                    help = TEST_HELP)
        }

        // No problem.
        Args(parserOf("-yx")).run {
            assertTrue(x)
            assertTrue(y)
        }

        // Attempting to give -y a parameter, "z", is treated as unrecognized option.
        shouldThrow<UnrecognizedOptionException> {
            Args(parserOf("-yz")).y
        }.run {
            assertEquals("unrecognized option '-z'", message)
        }

        // Unconsumed "z" again, but note that it triggers even if we don't look at y.
        shouldThrow<UnrecognizedOptionException> {
            Args(parserOf("-yz")).x
        }.run {
            assertEquals("unrecognized option '-z'", message)
        }

        // No problem again, this time with long opts.
        Args(parserOf("--why", "--ecks")).run {
            assertTrue(x)
            assertTrue(y)
        }

        // Attempting to give --why a parameter, "z" causes an error.
        shouldThrow<UnexpectedOptionArgumentException> {
            Args(parserOf("--why=z")).y
        }.run {
            assertEquals("option '--why' doesn't allow an argument", message)
        }

        // Unconsumed "z" again, but note that it triggers even if we don't look at y.
        shouldThrow<UnexpectedOptionArgumentException> {
            Args(parserOf("--why=z")).x
        }.run {
            assertEquals("option '--why' doesn't allow an argument", message)
        }
    }

    @Test
    fun testPositional_basic() {
        class Args(parser: ArgParser) {
            val flag by parser.flagging("-f", "--flag",
                    help = TEST_HELP)
            val store by parser.storing("-s", "--store",
                    help = TEST_HELP).default("DEFAULT")
            val sources by parser.positionalList("SOURCE",
                    help = TEST_HELP)
            val destination by parser.positional("DEST",
                    help = TEST_HELP)
        }

        Args(parserOf("foo", "bar", "baz", "quux")).run {
            assertFalse(flag)
            assertEquals("DEFAULT", store)
            assertEquals(listOf("foo", "bar", "baz"), sources)
            assertEquals("quux", destination)
        }

        Args(parserOf("-f", "foo", "bar", "baz", "quux")).run {
            assertTrue(flag)
            assertEquals("DEFAULT", store)
            assertEquals(listOf("foo", "bar", "baz"), sources)
            assertEquals("quux", destination)
        }

        Args(parserOf("-s", "foo", "bar", "baz", "quux")).run {
            assertFalse(flag)
            assertEquals("foo", store)
            assertEquals(listOf("bar", "baz"), sources)
            assertEquals("quux", destination)
        }

        Args(parserOf("-s", "foo", "bar", "-f", "baz", "quux")).run {
            assertTrue(flag)
            assertEquals("foo", store)
            assertEquals(listOf("bar", "baz"), sources)
            assertEquals("quux", destination)
        }

        // "--" disables option processing for all further arguments.
        // Note that "-f" is now considered a positional argument.
        Args(parserOf("-s", "foo", "--", "bar", "-f", "baz", "quux")).run {
            assertFalse(flag)
            assertEquals("foo", store)
            assertEquals(listOf("bar", "-f", "baz"), sources)
            assertEquals("quux", destination)
        }

        // "--" disables option processing for all further arguments.
        // Note that the second "--" is also considered a positional argument.
        Args(parserOf("-s", "foo", "--", "bar", "--", "-f", "baz", "quux")).run {
            assertFalse(flag)
            assertEquals("foo", store)
            assertEquals(listOf("bar", "--", "-f", "baz"), sources)
            assertEquals("quux", destination)
        }

        Args(parserOf("-s", "foo", "bar", "-f", "baz", "quux", mode = ArgParser.Mode.POSIX)).run {
            assertFalse(flag)
            assertEquals("foo", store)
            assertEquals(listOf("bar", "-f", "baz"), sources)
            assertEquals("quux", destination)
        }
    }

    @Test
    fun testPositional_withParser() {
        class Args(parser: ArgParser) {
            val flag by parser.flagging("-f", "--flag",
                    help = TEST_HELP)
            val store by parser.storing("-s", "--store",
                    help = TEST_HELP).default("DEFAULT")
            val start by parser.positionalList("START", 3..4,
                    help = TEST_HELP) { toInt() }
            val end by parser.positionalList("END", 3..5,
                    help = TEST_HELP) { toInt() }
        }

        shouldThrow<MissingRequiredPositionalArgumentException> {
            Args(parserOf("1", "2")).flag
        }.run {
            assertEquals("missing START operand", message)
        }

        shouldThrow<MissingRequiredPositionalArgumentException> {
            Args(parserOf("1", "2", "3", "4", "5")).flag
        }.run {
            assertEquals("missing END operand", message)
        }

        Args(parserOf("1", "2", "3", "4", "5", "6")).run {
            assertFalse(flag)
            assertEquals("DEFAULT", store)

            // end needs at least 3 args, so start only consumes 3
            assertEquals(listOf(1, 2, 3), start)
            assertEquals(listOf(4, 5, 6), end)
        }

        Args(parserOf("1", "2", "3", "4", "5", "6", "7")).run {
            assertFalse(flag)
            assertEquals("DEFAULT", store)

            // end only needs at 3 args, so start can consume 4
            assertEquals(listOf(1, 2, 3, 4), start)
            assertEquals(listOf(5, 6, 7), end)
        }

        Args(parserOf("1", "2", "3", "4", "5", "6", "7", "8")).run {
            assertFalse(flag)
            assertEquals("DEFAULT", store)

            // start can't consume more than 4, so end gets the rest.
            assertEquals(listOf(1, 2, 3, 4), start)
            assertEquals(listOf(5, 6, 7, 8), end)
        }

        Args(parserOf("1", "2", "3", "4", "5", "6", "7", "8", "9")).run {
            assertFalse(flag)
            assertEquals("DEFAULT", store)

            // once again, start can't consume more than 4, so end gets the rest.
            assertEquals(listOf(1, 2, 3, 4), start)
            assertEquals(listOf(5, 6, 7, 8, 9), end)
        }

        shouldThrow<UnexpectedPositionalArgumentException> {
            Args(parserOf("1", "2", "3", "4", "5", "6", "7", "8", "9", "10")).flag
        }.run {
            assertEquals("unexpected argument after END", message)
        }
    }

    @Test
    fun testCounting() {
        class Args(parser: ArgParser) {
            val verbosity by parser.counting("-v", "--verbose",
                    help = TEST_HELP)
        }

        Args(parserOf()).run {
            assertEquals(0, verbosity)
        }

        Args(parserOf("-v")).run {
            assertEquals(1, verbosity)
        }

        Args(parserOf("-v", "-v")).run {
            assertEquals(2, verbosity)
        }
    }

    @Test
    fun testHelp() {
        class Args(parser: ArgParser) {
            val dryRun by parser.flagging("-n", "--dry-run",
                    help = "don't do anything")
            val includes by parser.adding("-I", "--include",
                    help = "search in this directory for header files")
            val outDir by parser.storing("-o", "--output",
                    help = "directory in which all output should be generated")
            val verbosity by parser.counting("-v", "--verbose",
                    help = "increase verbosity")
            val sources by parser.positionalList("SOURCE",
                    help = "source file")
            val destination by parser.positional("DEST",
                    help = "destination file")
        }

        shouldThrow<ShowHelpException> {
            Args(parserOf("--help",
                    helpFormatter = DefaultHelpFormatter(
                            prologue = """
                            This is the prologue. Lorem ipsum dolor sit amet, consectetur adipiscing elit. Aliquam malesuada maximus eros. Fusce luctus risus eget quam consectetur, eu auctor est ullamcorper. Maecenas eget suscipit dui, sed sodales erat. Phasellus.
                            """,
                            epilogue = """
                            This is the epilogue. Lorem ipsum dolor sit amet, consectetur adipiscing elit. Donec vel tortor nunc. Sed eu massa sed turpis auctor faucibus. Donec vel pellentesque tortor. Ut ultrices tempus lectus fermentum vestibulum. Phasellus.
                            """))).dryRun
        }.run {
            val writer = StringWriter()
            printUserMessage(writer, "program_name", 60)
            val help = writer.toString()
            assertEquals(
                    """
usage: program_name [-h] [-n] [-I INCLUDE]... -o OUTPUT
                    [-v]... SOURCE... DEST


This is the prologue. Lorem ipsum dolor sit amet, consectetur
adipiscing elit. Aliquam malesuada maximus eros. Fusce
luctus risus eget quam consectetur, eu auctor est
ullamcorper. Maecenas eget suscipit dui, sed sodales erat.
Phasellus.


required arguments:
  -o OUTPUT,        directory in which all output should
  --output OUTPUT   be generated


optional arguments:
  -h, --help        show this help message and exit

  -n, --dry-run     don't do anything

  -I INCLUDE,       search in this directory for header
  --include INCLU   files
  DE

  -v, --verbose     increase verbosity


positional arguments:
  SOURCE            source file

  DEST              destination file



This is the epilogue. Lorem ipsum dolor sit amet, consectetur
adipiscing elit. Donec vel tortor nunc. Sed eu massa sed
turpis auctor faucibus. Donec vel pellentesque tortor. Ut
ultrices tempus lectus fermentum vestibulum. Phasellus.

""".trimStart(), help)
        }
    }

    @Test
    fun testImplicitLongFlagName() {
        class Args(parser: ArgParser) {
            val flag1 by parser.flagging(help = TEST_HELP)
            val flag2 by parser.flagging(help = TEST_HELP)
            val count by parser.counting(help = TEST_HELP)
            val store by parser.storing(help = TEST_HELP)
            val store_int by parser.storing(help = TEST_HELP) { toInt() }
            val adder by parser.adding(help = TEST_HELP)
            val int_adder by parser.adding(help = TEST_HELP) { toInt() }
            val int_set_adder by parser.adding(initialValue = mutableSetOf<Int>(), help = TEST_HELP) { toInt() }
            val positional by parser.positional(help = TEST_HELP)
            val positional_int by parser.positional(help = TEST_HELP) { toInt() }
            val positionalList by parser.positionalList(sizeRange = 2..2, help = TEST_HELP)
            val positionalList_int by parser.positionalList(sizeRange = 2..2, help = TEST_HELP) { toInt() }
        }

        Args(parserOf(
                "--flag1", "--count", "--count", "--store=hello", "--store-int=42",
                "--adder=foo", "--adder=bar",
                "--int-adder=2", "--int-adder=4", "--int-adder=6",
                "--int-set-adder=64", "--int-set-adder=128", "--int-set-adder=20",
                "1", "1", "2", "3", "5", "8"
                )).run {
            assertTrue(flag1)
            assertFalse(flag2)
            assertEquals(2, count)
            assertEquals("hello", store)
            assertEquals(42, store_int)
            assertEquals(listOf("foo", "bar"), adder)
            assertEquals(listOf(2, 4, 6), int_adder)
            assertEquals(setOf(20, 64, 128), int_set_adder)
            assertEquals("1", positional)
            assertEquals(1, positional_int)
            assertEquals(listOf("2", "3"), positionalList)
            assertEquals(listOf(5, 8), positionalList_int)
        }

        shouldThrow<MissingRequiredPositionalArgumentException> {
            Args(parserOf(
                    "13", "21", "34", "55", "89"
            )).run {
                assertFalse(flag1)
            }
        }.run {
            assertEquals("missing POSITIONALLIST_INT operand", message)
        }
    }

    // TODO: test auto-naming on positional args
    // TODO: test default on argument
    // TODO: test default on argumentList
    // TODO: test addValidator on argument
    // TODO: test addValidator on argumentList
}
